"""
backbonetools - create, modify and visualise input and output data for the esm backbone and run backbone
Copyright (C) 2020-2025 Leonie Plaga, David Huckebrink, Christine Nowak, Jan Mutke, Jonas Finke, Silke Johanndeiter, Sophie Pathe

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import subprocess as sp
from glob import glob
from multiprocessing.pool import ThreadPool
from pathlib import Path

from backbonetools.io.input import BackboneInput
from backbonetools.io.output import BackboneResult


def run_backbone(
    input_dir,
    input_file_gdx,
    output_dir,
    output_file=None,
    backbone_dir=None,
    **kwargs,
):
    """runs Backbone with the specified arguments

    Args:
        input_dir (str): Path to the input directory for Backbone to read configurations (e.g. `investInit.gms`)
        input_file_gdx (str): Path to the Backbone input file
        output_dir (str): Path to the output directory, where results shall be written
        output_file (str): Name of the Backbone result file
        backbone_dir (str, optional): Path to the Backbone Framework. Defaults to the path of the submodule in this repository.
    """

    # if not specified, derive from input_file_gdx
    if not output_file:
        output_file = Path(input_file_gdx).stem + "_result.gdx"

    # if not specified, set backbone_dir to submodule path
    if not backbone_dir:
        backbone_dir = Path(__file__).parent.parent.joinpath("backbone")

    # keyword arguments are parsed into backbone suitable form
    # e.g. dict(maxTotalCost=42) will be passed as "--maxTotalCost=42" to backbone
    keyword_args = [f"--{k}={v}" for k, v in kwargs.items()]

    # spaces in file names don't work through subprocesses
    contain_spaces = [
        " " in str(file_or_dir)
        for file_or_dir in [input_dir, input_file_gdx, output_dir, output_file]
    ]
    if any(contain_spaces):
        raise ValueError(
            "Passing paths with spaces via subprocess to Gams is not supported yet."
        )

    # absolute paths are better for gams
    input_dir = Path(input_dir).absolute().as_posix()
    input_file_gdx = Path(input_file_gdx).absolute().as_posix()
    output_dir = Path(output_dir).absolute().as_posix()

    Path(input_dir).mkdir(exist_ok=True)
    Path(output_dir).mkdir(exist_ok=True)

    process_output = sp.run(
        [
            "gams",
            "Backbone.gms",
            f"--input_dir={input_dir}",
            f"--input_file_gdx={input_file_gdx}",
            f"--output_dir={output_dir}",
            f"--output_file={output_file}",
            *keyword_args,
        ],
        cwd=backbone_dir,
        stdout=sp.PIPE,
        stderr=sp.PIPE,
        # text= True
    )

    # stdout, stderr = process_output.communicate()
    # write terminal output to a file
    open(f"{output_dir}/{Path(output_file).stem}_terminal.log", "w").write(
        process_output.stdout.decode()
    )
    open(f"{output_dir}/{Path(output_file).stem}_error.log", "w").write(
        process_output.stderr.decode()
    )
    print("finished", output_file)
    return Path(output_dir).joinpath(output_file).absolute()


def sensitivity(
    input_file_gdx,
    bb_input_dir,
    output_dir=None,
    parallel=True,
    param="CO2",
    percentages=[0.05, 0.1, 0.2, 0.4, 0.8],
):
    if param != "CO2":
        raise NotImplementedError(
            f"sensitivity not yet implemented for parameters other than {param}."
        )

    sensitivity_dir = Path(input_file_gdx).parent
    sensitivity_input_dir = f"{sensitivity_dir}/{param}_inputs"
    if not output_dir:
        sensitivity_output_dir = f"{sensitivity_dir}/{param}_outputs"
    else:
        sensitivity_output_dir = output_dir

    # create directories
    Path(sensitivity_input_dir).mkdir(exist_ok=True)
    Path(sensitivity_output_dir).mkdir(exist_ok=True)

    bb_in = BackboneInput(input_file_gdx)
    bb_opt_fn = bb_in._path.stem + f"_100{param}_result.gdx"

    # run first optimisation i.e. 100% of `param`
    bb_out_path = run_backbone(
        bb_input_dir,
        input_file_gdx=bb_in._path,
        output_dir=sensitivity_output_dir,
        output_file=bb_opt_fn,
    )

    # from scripts.results import BackboneResult
    bb_result = BackboneResult(bb_out_path)
    bb_result.r_emission()

    emission_lims = bb_result.r_emission()["Val"].values * percentages
    # percentages,emission_lims

    for percentage, em_lim in zip(percentages, emission_lims):
        new_db = bb_in.update_gams_parameter_in_db(
            "p_groupPolicyEmission",
            ["emission group", "emissionCap", "CO2"],
            value=em_lim,
        )
        out_fn = Path(bb_in._path).stem + f"_{percentage*100:02.0f}{param}"
        Path(sensitivity_input_dir).mkdir(exist_ok=True)
        new_db.export(f"{sensitivity_input_dir}/{out_fn}.gdx")

    input_files = list(glob(f"{sensitivity_input_dir}/*.gdx"))
    result_paths = []

    if parallel:
        threads = os.cpu_count() - 1
        with ThreadPool(threads) as pool:
            jobs = []
            for file in input_files:
                job = pool.apply_async(
                    run_backbone, (bb_input_dir, file, sensitivity_output_dir)
                )
                jobs.append(job)
            result_paths = [job.get() for job in jobs]
    else:
        for file in input_files:
            path = run_backbone(bb_input_dir, file, sensitivity_output_dir)
            result_paths.append(path)

    return result_paths



"""
backbonetools - create, modify and visualise input and output data for the esm backbone and run backbone
Copyright (C) 2020-2025 Leonie Plaga, David Huckebrink, Christine Nowak, Jan Mutke, Jonas Finke, Silke Johanndeiter, Sophie Pathe

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import subprocess as sp
import warnings
from functools import partial, reduce
from io import StringIO
from itertools import groupby
from pathlib import Path

import pandas as pd
import tsam.timeseriesaggregation as tsam
from gams import GamsDatabase, GamsException, GamsSetRecord, GamsWorkspace


class BackboneInput:
    def __init__(self, path):
        # this section is redundant with results.BackboneResult
        ws = GamsWorkspace(".")

        self._path = Path(path).absolute()

        if not self._path.exists():
            raise FileNotFoundError(self._path)

        self.gams_db = ws.add_database_from_gdx(self._path.as_posix())

        self.set_symbols_as_attribs()

    @property
    def input_symbols(self):
        if hasattr(self, "_symbols"):
            return self._symbols
        # get all symbols from the gams database
        # list comprehension of gams_db raises an exception at the last index
        symbols = []
        try:
            for p in self.gams_db:
                symbols.append(p.name)
        except GamsException as ge:
            if "out of range" in str(ge) and str(self.gams_db.number_symbols) in str(
                ge
            ):
                pass
            else:
                raise ge
        self._symbols = symbols
        return symbols

    def set_symbols_as_attribs(self):
        """
        Sets all input symbols as methods. Allows for easy access like
        `bb_input.p_gnu_io()` or `bb_input.ts_influx()`"""
        for symbol in self.input_symbols:
            setattr(self, symbol, partial(self.param_as_df_gdxdump, symbol))

    def param_as_df_gdxdump(
        self, symbol, encoding="1252", convert_time=True
    ) -> pd.DataFrame:
        """Use the 'gdxdump' GAMS utility via subprocess to convert a parameter into a pd.DataFrame.
        This is sometimes beneficial, to circumvent decoding errors
        """
        gdxdump = sp.run(
            ["gdxdump", self._path, "format", "csv", "symb", symbol], stdout=sp.PIPE
        )
        csv_data = gdxdump.stdout.decode(encoding=encoding)
        header = csv_data.partition("\n")[0]

        header = [x.strip('"') for x in header.split(",")]
        dtypes = [str if x != "Val" else float for x in header]
        dtypes = dict(zip(header, dtypes))
        df = pd.read_csv(StringIO(csv_data), dtype=dtypes, na_values="Eps")

        if df.empty:
            return df

        # converts the time column to int by omitting the first character
        try:
            if convert_time:
                if "t" in df.columns:
                    df["t"] = df["t"].apply(lambda t: int(t[1:]))
                else:
                    # for backbone inputs, the index is ["Dim1", "Dim2", ... , "Val"], but there might be a time column
                    first_row = df.loc[0]
                    is_time = first_row.apply(lambda x: str(x).startswith("t000"))

                    if any(is_time):
                        # any is necessary, otherwise *.idxmax()  simply returns first element
                        time_col = is_time.idxmax()  # should be something like "Dim4"
                        df[time_col] = df[time_col].apply(lambda t: int(t[1:]))
                        df.rename({time_col: "t"}, axis=1, inplace=True)
        except Exception as e:
            print(f"Error during time conversion for {symbol=}, {df=}")

        return df

    def update_gams_parameter_in_db(
        self,
        param_name: str,
        indices: list,
        value: float = None,
        apply_percentage: float = None,
    ) -> GamsDatabase:
        # note that you are always only working on the same instance of the database.
        # retrieve the parameter i.e. "p_groupPolicyEmission"
        parameter = self.gams_db.get_symbol(param_name)

        # retrieve the record from that parameter i.e. ["emission group", "emissionCap", "CO2"]
        try:
            record = parameter.find_record(indices)
        except GamsException as ge:
            if "Cannot find" in ge.value:
                # This is potentially dangerous since any index value can be added.
                # However, in the case of a bad index, GAMS should throw a DomainViolationError
                warnings.warn(
                    f"Warning: Caught 'GamsException: {ge.value}'.\n\tSetting {indices} to {value}"
                )
                record = parameter.add_record(indices)
            else:
                raise ge
        # add_record
        if not apply_percentage and type(value) is not None:
            if isinstance(record, GamsSetRecord):
                # sets only have text i.e. "Y"
                record.set_text(value)
            else:
                record.set_value(value)
        elif apply_percentage:
            current_val = record.get_value()
            record.set_value(current_val * apply_percentage)
        else:
            raise ValueError("one of 'value' and 'apply_percentage' must be passed!")
        return self.gams_db

    def aggregate_timeseries(
        self, export_path, hours_per_period=24 * 7, n_periods=3, err_indicators=None
    ):
        # err_indicators="print"|None
        if err_indicators not in ("print", None):
            raise NotImplementedError(
                f" Only one of ('print',None) is supported for `err_indicators`, but {err_indicators=} was passed."
            )

        ts_params = ["ts_influx", "ts_node", "ts_unit", "ts_cf"]

        ts_frames = []
        drop_frames = []
        # retrieve and transform timeseries data (not exaustive)
        for p in ts_params:
            p_df = self.param_as_df_gdxdump(p)
            if p_df.empty:
                drop_frames.append(p)
                continue
            # index of the time column
            time_col = list(p_df.columns).index("t")

            # all the previous columns (time is usually the penultimate column)
            index_cols = list(p_df.columns[:time_col])

            # transform from long to wide dataformat
            p_df = p_df.pivot(index="t", columns=index_cols, values="Val")

            # join column names from different levels and prepend param_name
            # these will come in handy, when writing aggregated data to the *.gdx
            # col is tuple, thus must be converted to a list to be mutable
            p_df.columns = ["|".join([p] + list(col)) for col in p_df.columns]

            ts_frames.append(p_df)
        # merge ts_frames to a single dataframe
        ts_df = reduce(
            lambda left, right: pd.merge(
                left, right, left_index=True, right_index=True, how="outer"
            ),
            ts_frames,
        )

        # create and set a datetimeindex
        dateindex = pd.date_range(start="2020-01-01 00:00", periods=8760, freq="h")
        ts_df.set_index(dateindex, inplace=True)
        ts_df.fillna(0, inplace=True)

        # create aggregation
        aggregation = tsam.TimeSeriesAggregation(
            ts_df,
            hoursPerPeriod=hours_per_period,
            noTypicalPeriods=n_periods,
            clusterMethod="adjacent_periods",
            representationMethod="distributionAndMinMaxRepresentation",
        )

        # calculate the typical periods
        typeriods = aggregation.createTypicalPeriods()

        if err_indicators == "print":
            print(aggregation.accuracyIndicators())

        # get the sequence and number of occurances
        period_sequence = [
            (k, sum(1 for i in g)) for k, g in groupby(aggregation.clusterOrder)
        ]

        # concatenate
        ts_input_data = pd.concat(
            [typeriods.loc[period_no, :] for period_no, _ in period_sequence]
        ).reset_index(drop=True)

        # write data to gdx
        for col_name in ts_input_data.columns:
            # recreate the indices from the column names
            bb_indices = col_name.split("|")

            for i, val in enumerate(ts_input_data[col_name].values):
                timestep = f"t{i + 1:06}"
                self.update_gams_parameter_in_db(
                    bb_indices[0], bb_indices[1:] + [timestep], val
                )

        # add samples to .gdx where nessesary
        samples_and_lengths = [
            {
                "name": f"s{i:03}",
                "length": hours_per_period,
                "weight": tup[1],
                "cluster_no": tup[0],
            }
            for i, tup in enumerate(period_sequence)
        ]
        sample_df = pd.DataFrame.from_records(samples_and_lengths, index="name")
        sample_names = sample_df.index.to_list()

        if "p_s_discountFactor" not in self.input_symbols:
            self.gams_db.add_parameter_dc("p_s_discountFactor", ["s"])
        for sample in sample_df.index:
            self.update_gams_parameter_in_db("p_s_discountFactor", [sample], value=1)
            # : ["sample",	"group"]
            self.update_gams_parameter_in_db(
                "sGroup", [sample, "emission group"], value="Y"
            )
            assert "emission group" in self.group().values

        if len(self.group().values) > 2:
            print(
                f"Warning: samples have been added for group 'emission group' but more are present in the .gdx:{self.group().values}"
            )

        # bind samples in gnss_bound
        # get nodes that have a state
        p_gn = self.p_gn()
        nodes_w_state = p_gn.query("Dim3=='energyStoredPerUnitOfState'")[
            ["Dim1", "Dim2"]
        ]
        # nodes_w_state.append([nodes_w_state]*(len(sample_names)-1))
        gnss_df = pd.concat([nodes_w_state] * len(sample_names))
        gnss_df.sort_values(by="Dim2", inplace=True)

        from_sample = []
        to_sample = []
        for i in range(len(gnss_df)):
            from_sample.append(sample_names[i % len(sample_names)])
            to_sample.append(sample_names[(i + 1) % len(sample_names)])

        gnss_df["from_sample"] = from_sample
        gnss_df["to_sample"] = to_sample

        # clear records in bounds, because they might be invalid post-aggregation
        gnss_bounds = self.gams_db.get_symbol("gnss_bound")
        gnss_bounds.clear()

        for gnss_indices in gnss_df.values:
            # print(gnss_indices)
            self.update_gams_parameter_in_db(
                "gnss_bound", list(gnss_indices), value="Y"
            )

        print("Dont forget to update your bb input configuration!!")
        for i, tup in enumerate(period_sequence):
            print(
                f"s{i:03} - cluster:{tup[0]} - weight:{tup[1]},\tlength: {hours_per_period}"
            )
        sample_df.to_csv(str(export_path) + "_sample_weights.csv")

        self.gams_db.export(export_path)
        pass



"""
backbonetools - create, modify and visualise input and output data for the esm backbone and run backbone
Copyright (C) 2020-2025 Leonie Plaga, David Huckebrink, Christine Nowak, Jan Mutke, Jonas Finke, Silke Johanndeiter, Sophie Pathe

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import subprocess as sp
from functools import partial
from io import StringIO
from multiprocessing.pool import ThreadPool
from pathlib import Path

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from gams import GamsException, GamsWorkspace
from matplotlib.colors import to_hex
from matplotlib.pyplot import get_cmap
from openpyxl import load_workbook


class BackboneResult:
    def __init__(self, path):
        ws = GamsWorkspace()
        self._path = Path(path).absolute().as_posix()
        try:
            self.gams_db = ws.add_database_from_gdx(self._path)
        except GamsException as ge:
            raise GamsException(f"Error when opening {self._path}").with_traceback(
                ge.__traceback__
            )
        except TypeError as te:
            raise TypeError(f"An error occured on {self._path}:").with_traceback(
                te.__traceback__
            )

        try:
            self.mMettings = self.gams_db.get_parameter("r_info_mSettings")
        except GamsException as ge:
            raise GamsException(
                f"Error retrieving 'r_info_mSettings' from {self._path}.\n"
                f"Please check if the file is a valid Backbone result file."
            ).with_traceback(ge.__traceback__)
        self.stepLengthInHours = self.mMettings.find_record(
            keys=["stepLengthInHours"]
        ).value
        # invest or schedule
        self.model_type = self.mMettings.first_record().keys[0]

        # allows easy access of results
        self.set_symbols_as_attribs()
        self.has_dummy_generation = not self.r_qGen_g().empty
        if self.has_dummy_generation:
            Warning(f"Model has dummy generation. ({Path(path).name})")

    def __hash__(self):
        # this is not a good hash implementation but suffices for the purpose of not loading the same file multiple times
        return hash(self.path.as_posix())

    def set_symbols_as_attribs(self):
        for symbol in self.symbols:
            # this allows to access result parameters by BackboneResult.r_emission()
            setattr(self, symbol, partial(self.param_as_df_gdxdump, symbol))

    @property
    def _backbone_path(self):
        """path to the backbone repository"""
        return Path(__file__).parent.parent.joinpath("backbone")

    @property
    def symbols(self):
        """list of all symbols in the {result_file}.gdx"""
        if hasattr(self, "_symbols"):
            return self._symbols
        # get all symbols from the gams database
        # list comprehension of gams_db raises an exception at the last index
        symbols = []
        try:
            for i, p in enumerate(self.gams_db):
                symbols.append(p.name)
        except GamsException as ge:
            if "out of range" in str(ge) and str(self.gams_db.number_symbols) in str(
                ge
            ):
                pass
            else:
                raise ge
        self._symbols = symbols
        return symbols

    def param_as_df(self, param_name, convert_time=True):
        gms_param = self.gams_db.get_parameter(param_name)
        data = [[*rec.keys, rec.value] for rec in gms_param]

        # domains have changed and now may contain sets
        domains = [x if isinstance(x, str) else x.name for x in gms_param.domains]

        columns = domains + ["Val"]
        param_df = pd.DataFrame(data=data, columns=columns)

        # converts the time column to int by omitting the first character
        if convert_time:
            if "t" in param_df.columns:
                param_df["t"] = param_df["t"].apply(lambda t: int(t[1:]))

        return param_df

    def param_as_df_gdxdump(self, param_name, encoding="1252", convert_time=True):
        """Use the 'gdxdump' GAMS utility via subprocess to convert a parameter into a pd.DataFrame.
        This is sometimes beneficial, to circumvent decoding errors
        """
        gdxdump = sp.run(
            ["gdxdump", self._path, "format", "csv", "symb", param_name], stdout=sp.PIPE
        )
        csv_data = gdxdump.stdout.decode(encoding=encoding)
        header = csv_data.partition("\n")[0]

        header = [x.strip('"') for x in header.split(",")]
        dtypes = [str if x != "Val" else float for x in header]
        dtypes = dict(zip(header, dtypes))
        df = pd.read_csv(StringIO(csv_data), dtype=dtypes, na_values="Eps")

        # converts the time column to int by omitting the first character
        if convert_time:
            if "t" in df.columns:
                df["t"] = df["t"].apply(lambda t: int(t[1:]))
        return df

    def eqn_as_df(self, eqn_name):
        gms_eqn = self.gams_db.get_equation(eqn_name)

        data = [
            [*rec.keys, rec.level, rec.marginal, rec.lower, rec.upper, rec.scale]
            for rec in gms_eqn
        ]
        columns = [x if isinstance(x, str) else x.name for x in gms_eqn.domains] + [
            "level",
            "marginal",
            "lower",
            "upper",
            "scale",
        ]

        eqn_df = pd.DataFrame(data=data, columns=columns)
        return eqn_df

    def set_as_df(self, set_name):
        gms_set = self.gams_db.get_set(set_name)
        data = [[*rec.keys] for rec in gms_set]
        domains = [x if isinstance(x, str) else x.name for x in gms_set.domains]

        set_df = pd.DataFrame(data=data, columns=domains)
        return set_df


class BackboneScenarioAnalysis:
    def __init__(self, results_dir: str = None, result_files: list = None) -> None:
        if results_dir and not result_files:
            abs_dir = Path(results_dir).resolve()
            unsorted_results = list(abs_dir.glob(r"*.gdx"))
            self.result_files = sorted(unsorted_results)
        elif not results_dir and result_files:
            self.result_files = sorted([Path(path) for path in result_files])
        else:
            raise ValueError(
                f"""One of the parameters 'results_dir' or 'result_files' must be passed.
                Received: {dict(results_dir=results_dir, result_files=result_files)}.
                """
            )

        self.bb_results = [
            BackboneResult(Path(file).as_posix()) for file in self.result_files
        ]

        # scenario names will be derived from the filenames
        self.scenarios = [
            Path(file).stem.replace("_", " ") for file in self.result_files
        ]

        # define colors as sequence:
        self.colors = self._get_n_hex_colors_from_cmap(
            "Spectral_r", len(self.result_files)
        )

        self.set_symbols_as_attribs()
        pass

    def _scenario_param(self, param_name, parallel=False):
        frames = []

        def _get_attr_df_and_set_name(result, param_name, scenario_name):
            frame = getattr(result, param_name)()
            frame["Scenario"] = scenario_name
            return frame

        if not parallel:
            for result, scenario in zip(self.bb_results, self.scenarios):
                frame = _get_attr_df_and_set_name(result, param_name, scenario)
                frames.append(frame)
        else:
            with ThreadPool(9) as p:

                param_name_list = [param_name] * len(self.bb_results)
                frames = p.starmap(
                    _get_attr_df_and_set_name,
                    zip(self.bb_results, param_name_list, self.scenarios),
                )
        return pd.concat(frames)

    def set_symbols_as_attribs(self):
        for symbol in self.bb_results[0].symbols:
            # this allows to access result parameters by BackboneResult.r_emission()
            setattr(self, symbol, partial(self._scenario_param, symbol))

    @staticmethod
    def _get_n_hex_colors_from_cmap(cmap_name, n):
        cmap = get_cmap(cmap_name)
        fracs = np.linspace(0, 1, n)
        return [to_hex(cmap(f)) for f in fracs]

    def installed_capacity(self) -> pd.DataFrame:
        frames = []
        for scen, result in zip(self.scenarios, self.bb_results):
            installed_capacity = (
                result.r_invest_unitCapacity_gnu()
                .groupby("unit")
                .sum(numeric_only=True)
            )

            installed_capacity["Scenario"] = scen

            frames.append(installed_capacity)

        scenario_df = pd.concat(frames)
        scenario_df = scenario_df.reset_index()
        return scenario_df

    def plot_installed_capacity(self, colors=None):
        if colors is None:
            colors = self.colors
        scenario_df = self.installed_capacity()
        fig = px.bar(
            scenario_df,
            x="unit",
            y="Val",
            color="Scenario",
            barmode="group",
            color_discrete_sequence=colors,
        )

        return fig

    def system_cost_composition(
        self,
    ) -> pd.DataFrame:
        frames = []
        for scen, result in zip(self.scenarios, self.bb_results):
            # print(scen, result._path)
            invest_cost = result.r_cost_unitInvestmentCost_gnu()
            invest_cost["Category"] = "Invest"
            invest_cost_sum = invest_cost.groupby("Category").sum(numeric_only=True)

            transmission_cost = result.r_cost_linkInvestmentCost_gnn()
            transmission_cost["Category"] = "Grid expansion"
            transmission_cost_sum = transmission_cost.groupby("Category").sum(
                numeric_only=True
            )

            fuel_emission_cost = result.r_cost_unitFuelEmissionCost_u()
            fuel_emission_cost["Category"] = fuel_emission_cost["node"].str.capitalize()
            fuel_emission_cost_sum = fuel_emission_cost.groupby("Category").sum(
                numeric_only=True
            )

            fom_cost = result.r_cost_unitFOMCost_gnu()
            fom_cost["Category"] = "FOM"
            fom_cost_sum = fom_cost.groupby("Category").sum(numeric_only=True)

            vom_cost = result.r_cost_unitVOMCost_gnu()
            vom_cost["Category"] = "VOM"
            vom_cost_sum = vom_cost.groupby("Category").sum(numeric_only=True)

            cost_df = pd.concat(
                [
                    invest_cost_sum,
                    fuel_emission_cost_sum,
                    fom_cost_sum,
                    vom_cost_sum,
                    transmission_cost_sum,
                ]
            )
            cost_df["Scenario"] = scen
            frames.append(cost_df)
        scenario_df = pd.concat(frames)
        scenario_df = scenario_df.reset_index()
        return scenario_df

    def plot_system_cost_composition(self):
        scenario_df = self.system_cost_composition()
        fig = px.bar(scenario_df, x="Scenario", y="Val", color="Category", width=700)
        fig.update_yaxes({"title": "annual system cost [M€]"})
        fig.update_layout(legend={"traceorder": "reversed"})
        # fig.update_traces(dict(width=[0.6] * 8))
        return fig

    def energy_provided_per_unitType(
        self, unitType_mapping: dict = None
    ) -> pd.DataFrame:
        frames = []
        for scen, result in zip(self.scenarios, self.bb_results):
            sample_df = result.r_gen_gnu()
            grouped_df = sample_df.groupby(["grid", "unit"]).sum(numeric_only=True)
            grouped_df["Scenario"] = scen
            frames.append(grouped_df)
        scenario_df = pd.concat(frames)
        return scenario_df

    def plot_energy_per_unitType(self, colors=None):
        if colors is None:
            colors = self.colors
        fig = go.Figure()

        scenario_df = self.energy_provided_per_unitType().reset_index()

        for i, scen in enumerate(self.scenarios):
            data = scenario_df.loc[scenario_df["Scenario"] == scen, :]
            x = []
            x.append(data["grid"].values)
            x.append(data["unit"].values)
            color = colors[i % len(colors)]
            fig.add_bar(x=x, y=data["Val"], name=scen, marker={"color": color})
        # fig.update_traces(colors = px.colors.sequential.RdBu[:4])
        fig.update_yaxes(title="provided energy by unit type [MWh]")
        # fig.write_image(f"{results_dir}/figures/energy_generation_by_type.svg")
        # fig.write_image(f"{results_dir}/figures/energy_generation_by_type.pdf")
        return fig

    def plot_energy_per_unitType_stacked(self, colors=None, u_name_filter=None):
        fig = go.Figure()
        scenario_df = self.energy_provided_per_unitType().reset_index()

        if u_name_filter:
            # for a birds eye view, this can be helpful for condensing unit_names
            # i.e. u_name_filter=lambda x: x.split(" ",2)[-1]
            # "DE0 0 coal" -> coal
            # "FR0 1 coal" -> coal
            scenario_df["unit"] = scenario_df["unit"].apply(u_name_filter)

        utypes = scenario_df["unit"].unique()

        for i, utype in enumerate(utypes):
            # select and aggregate data
            data = scenario_df.loc[scenario_df["unit"] == utype, :]
            data = data.groupby(["grid", "unit", "Scenario"]).sum().reset_index()

            # generate double index
            x = []
            x.append(data["grid"].values)
            x.append(data["Scenario"].values)
            color = colors[i % len(colors)]
            fig.add_bar(x=x, y=data["Val"], name=utype, marker={"color": color})
        fig.update_yaxes(title="provided energy by unit type [MWh]")
        fig.update_layout(barmode="relative")
        return fig

    def plot_priceDurationCurve(self, nodes=[], stylelist=[], colorlist=None):
        # Plot Price Duration Curve by scenarios over all or predefined nodes in the electric grid
        result = self.bb_results[0]
        marginal = result.r_balance_marginalValue_gnft()
        # extract all electric nodes from first scenario if non are given
        # grid choice has to be adapted or dropped in the code if desired
        if not nodes:
            nodes = marginal[marginal["grid"] == "elec"]["node"].drop_duplicates()
        ax_list = []
        # define list for graph's line style
        if not stylelist:
            for x in range(len(self.scenarios)):
                for element in ["-", "--", "-.", ":"]:
                    stylelist.append(element)
        # create plots by node over scenarios
        for n in nodes:
            scen_dict = {}
            for s, r in zip(self.scenarios, self.bb_results):
                marginal = r.r_balance_marginalValue_gnft()
                marginal = marginal[marginal["node"] == str(n)].sort_values("Val")
                marginal["Val"] = marginal["Val"] * -1
                price = marginal["Val"].tolist()
                scen_dict.update({str(s): price})
                t = [np.arange(0, len(price), 1)]
            df = pd.DataFrame(scen_dict, index=t)
            ax = df.plot.line(title=str(n), lw=2, style=stylelist, color=colorlist)
            ax.set_xticks([])
            ax.set_yticks(
                range(
                    min(0, int(min(price))),
                    int(max(price)),
                    round((int(max(price)) - min(0, int(min(price)))) / 10),
                )
            )
            ax.set_xlabel("hours sorted")
            ax.set_ylabel("€/MWh")
            ax.legend()
            ax_list.append(ax)
        return ax_list

    # Plot Price Duration Curve by all or predefined electric nodes over scenario
    def plot_priceDurationCurve2(self, nodes=[], stylelist=[], colorlist=None):
        result = self.bb_results[0]
        marginal = result.r_balance_marginalValue_gnft()
        # extract all electric nodes from first scenario if non are given
        # grid choice has to be adapted or dropped in the code if desired
        if not nodes:
            nodes = (
                marginal[marginal["grid"] == "elec"]["node"]
                .drop_duplicates()
                .transpose()
                .tolist()
            )
        scen_dict = {}
        ax_list = []
        # define list for graph's line style
        if not stylelist:
            for x in range(len(nodes)):
                for element in ["-", "--", "-.", ":"]:
                    stylelist.append(element)
        # create plots by nodes over scenarios
        for s, r in zip(self.scenarios, self.bb_results):
            for n in nodes:
                # result=self.bb_results[0]
                marginal = r.r_balance_marginalValue_gnft()
                marginal = marginal[marginal["node"] == str(n)].sort_values("Val")
                marginal["Val"] = marginal["Val"] * -1
                price = marginal["Val"].tolist()
                t = [np.arange(0, len(price), 1)]
                scen_dict.update({str(n): price})
                df = pd.DataFrame(scen_dict, index=t)
            ax = df.plot.line(title=str(s), style=stylelist, lw=2, color=colorlist)
            ax.set_xticks([])
            ax.set_yticks(
                range(
                    min(0, int(min(price))),
                    int(max(price)),
                    round((int(max(price)) - min(0, int(min(price)))) / 10),
                )
            )
            ax.set_xlabel("hours sorted")
            ax.set_ylabel("€/MWh")
            ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
            ax_list.append(ax)
        return ax_list


def invest2schedule(invest_xlsx, result_gdx, schedule_xlsx_path=None):

    # # Invest 2 Schedule
    # ### read an investment result and its' input file, create a new input file

    bb_input_path = invest_xlsx
    bb_output_path = result_gdx

    if schedule_xlsx_path is None:
        schedule_xlsx_path = Path(bb_input_path).stem + "_schedule.xlsx"
        schedule_xlsx_path = Path(bb_input_path).parent.joinpath(schedule_xlsx_path)

    result = BackboneResult(bb_output_path)
    invested_capacities = result.r_invest_unitCapacity_gnu().dropna()

    wb = load_workbook(bb_input_path)
    p_gnu_io = wb["p_gnu_io"]
    p_gnu_io_values = list(p_gnu_io.values)
    p_gnu_io_df = pd.DataFrame(p_gnu_io_values[1:], columns=p_gnu_io_values[0])

    # we only put cost data, where input/output == "output"
    idx_output = p_gnu_io_df["input output"] == "output"

    output_df = p_gnu_io_df.loc[idx_output, :]
    int_idx = invested_capacities["unit"].apply(
        lambda x: output_df.index[output_df["unit"] == x][0]
    )

    # add investments to previously existing
    invest_rows = output_df.loc[int_idx, :]

    # replace eps with 0 to enable addition
    invest_rows = invest_rows.replace("eps", 0)
    invest_rows["capacity"] += invested_capacities["Val"].values

    # replace 0 with eps for backbone
    invest_rows = invest_rows.replace(0, "eps")

    # overwrite values in p_gnu_io frame
    # note that index of p_gnu_io_df and output_df is consistent
    p_gnu_io_df.loc[int_idx, "capacity"] = invest_rows["capacity"]

    p_gnu_io.delete_rows(idx=2, amount=9999)
    for tup in p_gnu_io_df.itertuples(index=False):
        # bb_indices = col_name.split("|")
        p_gnu_io.append([*tup])

    # disable investments: set max unit count to 0
    p_unit = wb["p_unit"]
    p_unit_values = list(p_unit.values)
    p_unit_df = pd.DataFrame(p_unit_values[1:], columns=p_unit_values[0])
    p_unit_df["maxUnitCount"] = 0

    p_unit.delete_rows(idx=2, amount=9999)
    for tup in p_unit_df.itertuples(index=False):
        p_unit.append([*tup])

    # Handle storages -TODO
    # Update upper limit for storages in p_gnBoundaryproperties

    storage_investments = invested_capacities.query("grid == 'storage'")
    if not storage_investments.empty:
        raise NotImplementedError(
            "Storages have been invested, but conversion to schedule is not yet implemented!"
        )

    # add investments in p_gnn
    grid_investments = result.r_invest_transferCapacity_gnn()
    grid_investments = grid_investments.set_index(["grid", "node", "node_1"])

    p_gnn_sheet = wb["p_gnn"]
    p_gnn_values = list(p_gnn_sheet.values)
    p_gnn_df = pd.DataFrame(p_gnn_values[1:], columns=p_gnn_values[0])

    # update index to enable easy merging of values
    p_gnn_df = p_gnn_df.set_index(["grid", "from node", "to node"])
    p_gnn_df.loc[grid_investments.index, "transferCap"] += grid_investments["Val"]

    # disable further investments
    p_gnn_df["transferCapInvLimit"] = "eps"

    # overwrite the sheet with updated values
    p_gnn_sheet.delete_rows(idx=2, amount=9999)
    p_gnn_df = p_gnn_df.reset_index()
    for tup in p_gnn_df.itertuples(index=False):
        p_gnn_sheet.append([*tup])

    wb.save(schedule_xlsx_path)
    return schedule_xlsx_path


if __name__ == "__main__":
    result = BackboneResult(
        "EURO_heat_retrofit_MICQP_bioShareLimrcp26_coldno_Cap_gFix_no_rtrft_results.gdx"
    )
    print(result.symbols)
    pass
